/*
 * Distributed under GPLv3
 * Copyright (C) 2011 Zhangxiaowen Gong
 * All right reserved
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * * Neither the name of the copyright holders nor the names of
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * \file DataBlock.h
 * \brief Header file for the data block base class
 *
 * \author Zhangxiaowen Gong
 * \date 2011/3/19
 */

#ifndef DATA_HPP_
#define DATA_HPP_

extern "C" {
#include "intc.h"
}

#include <stdint.h>
#include <cstddef>
#include "Utility.hpp"

/**
 * \defgroup data_block Data block
 * \{
 */

/**
 * \brief FreeRTOS hook function that executes when the system idle thread runs
 *
 * \see http://www.freertos.org/
 */
extern "C" void vApplicationIdleHook(void);

class CDataControl;

/**
 * \class CData
 *
 * \brief Base class that describes data blocks used by system I/O
 *
 * The data blocks are implemented as link list nodes. They can be queued either
 * in a system I/O stream, e.g. the CUSART sending queue, or in a queue waiting to free
 * the allocated memory. Generally, each system I/O has its own type of data block that
 * is inherited from CData
 *
 * <h2>Sample code:</h2>
 * \par Sample child class
 * \code
 *
 * class CSystemIO; // some system I/O
 *
 * class CChildData: public CData{
 * private:
 * 	uint8_t *m_data; // pointer to data that will be output via the system I/O
 * 	uint8_t m_len; // length of the block
 * public:
 * 	CChildData(uint8_t *data, uint8_t len){ // initialize the members
 * 		m_data = data;
 * 		m_len = len;
 * 	}
 * 	void free_sub_buffer(void);
 * 	friend class CSystemIO; // allow the IO class access private members in the data class
 * };
 *
 * \endcode
 *
 * \par sample member function that frees the data attached to the object
 * \code
 *
 * void CChildData::free_sub_buffer(void){
 * 	// if the data are allocated from the heap, then we should free it when the CData object is being freed
 * 	m_sub_buffer = m_data; // just simply make m_pucSubBlock point to the allocated data
 * }
 *
 * \endcode
 */

/**
 * \typedef pCData
 *
 * \brief Pointer type of CData
 */
typedef class CData {
protected:
	CData *m_next; ///< pointer to the next node in the queue
	uint8_t *m_sub_buffer; ///< pointer to the data that will be freed together with the CData object

public:
	/**
	 * \brief Constructor
	 *
	 * The constructor initialize the pointers for both next node and the sub-block to NULL
	 */
	CData(void) {
		m_next = NULL;
		m_sub_buffer = NULL;
	}

	friend class CDataControl;
}*pCData;

/**
 * \class CDataControl
 *
 * \brief The class manages the CData objects that will be freed
 *
 * When private member function CDataControl::free_all() is called, all CData objects
 * in the queue are freed
 *
 * <h2>Sample code:</h2>
 * \par Initialize the object
 * \code
 *
 * CDataControl *data_control = CDataControl::instance();
 *
 * \endcode
 *
 * \par Add a CData object to the queue
 * \code
 *
 * CData *data = new CData(); // a CData object
 * data_control -> wait_for_free(data); // tell the system to free the object when being idle
 *
 * \endcode
 */

/**
 * \typedef pCDataControl
 *
 * \brief Pointer type of CDataControl
 */
typedef class CDataControl {
private:
	CData *m_head; ///< pointer to the head of the free list
	CData *m_tail; ///< pointer to the tail of the free list
	volatile uint32_t m_len; ///< the length of the free list
	static CDataControl *m_instance; ///< pointer to the only instance that is able to be created

	/**
	 * \brief Constructor with no parameters
	 *
	 * Initialize the head and the tail of the free list
	 */
	CDataControl(void) {
		m_head = new CData();
		m_head->m_next = NULL;
		m_tail = m_head;
		m_len = 0;
	}

public:
	/**
	 * \brief Singleton implementation
	 *
	 * \return The pointer to the only instance that can be created
	 */
	static CDataControl *instance(void) {
		if (!m_instance)
			m_instance = new CDataControl();
		return m_instance;
	}

	/**
	 * \brief Destructor
	 *
	 * The destructor calls private member function CDataControl::free_all() to free
	 * all the objects in the free list before the CDataControl object is destroyed.
	 */
	~CDataControl(void) {
		free_all();
		m_instance = NULL;
	}

	/**
	 * \brief Add a CData object to the free list
	 *
	 * \param pData pointer to the CData object that is added to the free list
	 *
	 * \return the number of CData objects in the free list after the process is done
	 */
	uint32_t wait_for_free(CData *pData);

	/**
	 * \brief The function frees all the CData objects queued in the free list
	 *
	 * The function is called when FreeRTOS's system idle thread runs; thus, the CData objects
	 * in the free list are freed only if the system is idle. This ensures the performance of the
	 * system.
	 */
	void free_all(void);

	volatile uint32_t get_len(void) {
		return m_len;
	}
}*pCDataControl;

/**
 * \}
 */
#endif
